#include "BranchAndBound.hpp"
#include <limits>
#include <algorithm>
#include <cmath>

using namespace std;

BranchAndBound::BranchAndBound(Simplex root, const vector<bool>& integer):
	optimalSolution(root),//FIXME
	forceInteger(integer),
	upperBound(numeric_limits<double>::infinity()),
	done(false)
{
	activeNodes.push(root);
	Simplex& r = activeNodes.top();
	r.solve();
	lowerBoundStack.push(r.currentObjective());
}

void BranchAndBound::solve(double tolerance)
{
	/*
	 * NOTE: At the moment we're doing a depth-first search. This is a much better idea than doing a best-first search because it gets to bound, not just branch.
	 */

	if(done) return;
	done = true;
	
	feasible = false;
	
	while(!activeNodes.empty()){
		Simplex& curr = activeNodes.top();
		unsigned firstInfeasible;//passed by ref to feasible(), returns var to branch on.
		if(isFeasible(curr, tolerance, firstInfeasible)) {
			//IP Feasible, curr could possibly be the optimal solution
			if(curr.currentObjective() < upperBound) {
				//Call "optimal" if best so far
				optimalSolution = Simplex(curr);
				feasible = true;
				upperBound = curr.currentObjective();
			}
			removeCurrentNode();
		} else if(curr.currentObjective() >= upperBound || curr.provedInfeasible()) {
			//Curr and its children cannot be part of the optimal solution
			removeCurrentNode();
		} else {
			
			//A descendant of curr could possibly be the optimal solution
			//TODO: Slow copy-ctors. Make a data structure to take advantage of C++0X's move ctors when it comes out, for now never mind.
			Simplex left(curr);
			Simplex right(curr);

			left.addGEConstraint(firstInfeasible, ceil(left.getX().at(left.getInvVars().at(firstInfeasible))));
			left.solve();
			right.addLEConstraint(firstInfeasible, floor(right.getX().at(right.getInvVars().at(firstInfeasible))));
			right.solve();

			//Tree updates
			removeCurrentNode();
			if(lowerBoundStack.empty() || lowerBoundStack.top() >= left.currentObjective())
				lowerBoundStack.push(left.currentObjective());
			if(lowerBoundStack.top() >= right.currentObjective())
				lowerBoundStack.push(right.currentObjective());
			activeNodes.push(left);
			activeNodes.push(right);
		}
	}
}

void BranchAndBound::removeCurrentNode()
{
	if(lowerBoundStack.top() == activeNodes.top().currentObjective())
		lowerBoundStack.pop();
	
	activeNodes.pop();
}

bool BranchAndBound::isFeasible(const Simplex& s, double tolerance, unsigned& firstInfeasible)
{
	//TODO: Abstract away. Feasibility testing *is* a good time to find a branching candidate, though.
	if(s.provedInfeasible()) return false;
	
	const vector<int>& vars = s.getVariables();
	const vector<double>& x = s.getX();
	{
	vector<int>::const_iterator vIt = vars.begin();
	for(vector<double>::const_iterator xIt = x.begin(); xIt != x.end(); ++xIt, ++vIt) {
		if(*vIt < forceInteger.size() && forceInteger[*vIt] && min(abs(*xIt-floor(*xIt)), abs(*xIt-ceil(*xIt))) > tolerance) {
			firstInfeasible = *vIt;
			return false;
		}
	}
	}
	return true;
}
