#include "Simplex.hpp"
#include <limits>
#include <algorithm>
#include <functional>
#include <cmath>

using namespace std;

Simplex::Simplex(vector<vector<double> > inConstraints, vector<double> inCosts, vector<double> rightHandSide):
	constraints(vector<vector<double> >(inConstraints.size()+rightHandSide.size(), vector<double>(rightHandSide.size(), 0))),
	costs(inCosts),
	inverseB(vector<vector<double> >(rightHandSide.size(), vector<double>(rightHandSide.size(), 0))),
	x(vector<double>(rightHandSide.size())),
	variables(vector<int>(inConstraints.size()+rightHandSide.size(), 0)),
	invVariables(inConstraints.size()),
	totalVariables(inConstraints.size()),
	objective(0),
	optimal(false),
	done(false)
{
	for(int i = 0; i < x.size(); ++i) {
		if(rightHandSide.at(i) >= 0) {
			x.at(i) = rightHandSide.at(i);
			inverseB.at(i).at(i) = 1;
			constraints.at(i).at(i) = 1;
		} else {
			x.at(i) = -rightHandSide.at(i);
			inverseB.at(i).at(i) = -1;
			constraints.at(i).at(i) = -1;
		}
	}//x, basis, inverse basis initialise
	copy(inConstraints.begin(), inConstraints.end(), constraints.begin()+inverseB.size());
	//variables outside the basis
	for(int i = 0; i < invVariables.size(); ++i) {
		invVariables.at(i) = i + rightHandSide.size();
	}
	{
	int i=0;
	vector<int>::iterator j;
	for(j = variables.begin()+rightHandSide.size(); j != variables.end(); ++j)
		*j = i++;
	edgeArtificial = 0;
	for(j = variables.begin(); i != variables.size(); ++j, ++i)
		*j = --edgeArtificial;
	}//Regular variables out of the basis, artificials in.
}

void Simplex::solve(double epsilon)
{
	if(done) return;
	done = true;
	
	//TODO: test for artificials in the basis to check if we need to phase1. This works for the moment.
	
	bool needPhase1 = false;
	vector<double> fakeCosts(variables.size());
	for(unsigned i = 0; i < variables.size(); ++i) {
		if (variables[i] < 0 && i < x.size() && x[i] > epsilon) {
			needPhase1 = true;
			fakeCosts[i] = 1;
		}
	}
	if(needPhase1) realSolve(fakeCosts, false, epsilon);
	//Don't need to do phase 1 if we don't need to do phase 1.
	
	{
	vector<int>::const_iterator i = variables.begin();
	for(vector<double>::const_iterator j=x.begin(); j != x.end(); ++i, ++j) {
		if(*i < 0 && *j > epsilon) {
			//checks for artificials still in the basis.
			objective = numeric_limits<double>::infinity();
			return;
		}
	}
	}
	vector<double> realCosts(constraints.size());
	for(int i = 0; i < realCosts.size(); ++i) {
		if(variables.at(i) < 0)
			realCosts.at(i) = 0;//TODO: find out if this is correct
		else
			realCosts.at(i) = costs.at(variables.at(i));
	}//TODO: make this an in-place shuffle.

	optimal=true;//Say we're optimal even if unbounded.
	//Phase 2:
	if(realSolve(realCosts, true, epsilon)) {
		objective = -numeric_limits<double>::infinity();
		//Unbounded
	} else {
		objective = dotProduct(realCosts, x);
		//Optimal (properly), let objective = c*x
	}
}

bool Simplex::realSolve(vector<double>& costs, bool phase2, double epsilon)
{
	while(true) {
		unsigned enteringVariable;
		if(!findEV(enteringVariable, costs, epsilon))
			return false; //optimal
		unsigned leavingVariable;
		vector<double> decreases(inverseB.size(), 0);
		if(!findLV(leavingVariable, decreases, enteringVariable, phase2))
			return true;//unbounded
		updateSolution(leavingVariable, decreases);
		updateData(enteringVariable, leavingVariable, costs);
	}
}

bool Simplex::findEV(unsigned& enteringVariable, const vector<double> & costs, double epsilon)
{
	/*
	Minimum reduced cost algorithm. Don't know if we need "temp", but it's good enough.
	The second half can be cleaned up, made more functional, but this is a bit more efficient and not too opaque.
	Note: This function is probably a one-liner with Eigen... Should really port over in the future.
	*/

	vector<double> temp(inverseB.size(), 0);
	//inverse B now row-wise. So:
	//for i in rows of b, entries of cN:
		//for entries of temp, colB:
			//temp += cN*colB
	{
	vector<vector<double> >::const_iterator invBRow = inverseB.begin();
	for(vector<double>::const_iterator costIt = costs.begin(); invBRow != inverseB.end(); ++invBRow, ++costIt) {
		{
		vector<double>::const_iterator invBEntry = invBRow->begin();
		for(vector<double>::iterator tempIt = temp.begin(); tempIt != temp.end(); ++tempIt, ++ invBEntry)
			(*tempIt) += (*invBEntry)*(*costIt);
		}
	}
	}
	/*
	{
	vector<vector<double> >::const_iterator invBCol = inverseB.begin();
	for(vector<double>::iterator i = temp.begin(); i != temp.end(); ++invBCol , ++i){
		*i = dotProduct(costs, *invBCol);
	}
	}
	*/
	//temp = c_B * B^{-1}, row vector.

	enteringVariable = -1;
	double reducedCostOfEV = numeric_limits<double>::infinity();
	//Reduced cost = c_N - c_B * B^{-1} * N, we just want the index of the least element, and whether it's negative.
	
	{
	vector<vector<double> >::iterator nonbasisCol = constraints.begin()+inverseB.size();
	for(vector<double>::const_iterator nbCost = costs.begin()+inverseB.size(); nonbasisCol != constraints.end(); ++nonbasisCol, ++nbCost) {
		double currentRC = *nbCost - dotProduct(temp, *nonbasisCol);
		if(currentRC < reducedCostOfEV) {
			enteringVariable = nonbasisCol-constraints.begin();
			reducedCostOfEV = currentRC;
		}
	}
	}
	return reducedCostOfEV  < -epsilon;
}

bool Simplex::findLV(unsigned& leavingVariable, vector<double> & rateDecrease, unsigned enteringVariable, bool phase2)
{
	/*
	{
	vector<vector<double> >::const_iterator invBCol = inverseB.begin();
	for(vector<double>::const_iterator evColEntry = constraints.at(enteringVariable).begin(); evColEntry != constraints.at(enteringVariable).end(); ++evColEntry, ++invBCol) {
		vector<double>::iterator invBAsEntry = rateDecrease.begin();
		for(vector<double>::const_iterator invBEntry = invBCol->begin(); invBEntry != invBCol->end(); ++invBEntry, ++invBAsEntry) {
			(*invBAsEntry) += (*evColEntry) * (*invBEntry);
		}
	}
	}
	*/
	{
	vector<vector<double> >::const_iterator invBRow = inverseB.begin();
	for(vector<double>::iterator invBAsEntry = rateDecrease.begin(); invBAsEntry != rateDecrease.end(); ++invBAsEntry, ++invBRow)
		(*invBAsEntry) = dotProduct(*invBRow, constraints.at(enteringVariable));
	}
		
	//B^{-1}*A_s set up.
	/*
	rateDecrease = inverseB*nonBasis.col(enteringVariable); One line in Eigen again...
	Is ugly because inverseB is stored column-wise and I'm trying not to iterate row-wise.
	*/
	double minRatio = numeric_limits<double>::infinity();
	
	unsigned vectorLength = rateDecrease.size();
	//So we don't have to keep looking these up.
	{
	vector<int>::const_iterator varEntry = variables.begin();
	for(vector<double>::iterator invBAsEntry = rateDecrease.begin(), xEntry = x.begin(); invBAsEntry != rateDecrease.end(); ++invBAsEntry, ++varEntry, ++xEntry) {
		if(!phase2 || ((*varEntry) >= 0)) {
			if((*invBAsEntry) > 0) {
				double ratio = (*xEntry)/(*invBAsEntry);
				if(ratio < minRatio) {
					minRatio = ratio;
					leavingVariable = invBAsEntry-rateDecrease.begin();
				}
			}
		} else if ((*invBAsEntry) != 0) {
			//Phase 2, artificial in basis, remove it sometimes.
			*invBAsEntry = numeric_limits<double>::infinity();
			leavingVariable = invBAsEntry-rateDecrease.begin();
			return true;
			/*
			We remove it if it's going to change, which is a bad thing. EV
			enters at value 0. Change rate-change so the pivot does nothing.
			If (*i) == 0, this artificial is independent of the EV, so we'd
			be stopping good moves if we did this then.
			*/
		}
	}
	}
	return minRatio != numeric_limits<double>::infinity();
}

void Simplex::updateSolution(unsigned leavingVariable, const vector<double>& rateDecrease)
{
	/* "Gauss-Jordan pivot" on M = [inverseB, x, rateDecrease]:
	 * M.row(i) -= M.row(LV)*rateDecrease(i)/rateDecrease(LV) for i != LV,
	 * M.row(LV) /= rateDecrease(LV).
	 * Column-wise B-inverse storage makes it interesting.
	 */
	double rateDecreaseLV = rateDecrease.at(leavingVariable);
	
	{
	unsigned l=0;
	vector<double>::const_iterator thisRateDecrease = rateDecrease.begin();
	vector<vector<double> >::iterator invBRow = inverseB.begin();
	
	//Adjusting inverse B:
	//Rows 0:LV-1
	for(; l !=leavingVariable; ++invBRow, ++l, ++thisRateDecrease) {
		vector<double>::iterator invBEntry = invBRow->begin();
		double dec = *thisRateDecrease;
		for(vector<double>::const_iterator lvRowEntry = inverseB.at(leavingVariable).begin();
		    invBEntry != invBRow->end();
		    ++invBEntry, ++lvRowEntry)
			(*invBEntry) -= dec * (*lvRowEntry)/rateDecreaseLV;
	}
	//Row LV
	for(vector<double>::iterator lvInvBEntry = invBRow->begin(); lvInvBEntry != invBRow->end(); ++lvInvBEntry)
		(*lvInvBEntry) /= rateDecreaseLV;
	//Rows LV+1:End
	for(++thisRateDecrease, ++invBRow; invBRow != inverseB.end(); ++invBRow, ++thisRateDecrease) {
		vector<double>::iterator invBEntry = invBRow->begin();
		double dec = *thisRateDecrease;
		for(vector<double>::const_iterator lvRowEntry = inverseB.at(leavingVariable).begin();
		    invBEntry != invBRow->end();
		    ++invBEntry, ++lvRowEntry)
			(*invBEntry) -= dec * (*lvRowEntry);
	}
	}

	unsigned l=0;
	double xAtLV = x.at(leavingVariable)/rateDecreaseLV;
	vector<double>::const_iterator bInvAsEntry = rateDecrease.begin();
	vector<double>::iterator       xEntry = x.begin();
	
	//Adjusting x:
	//Rows 0:LV-1
	for(; l != leavingVariable; ++bInvAsEntry, ++xEntry, l++)
		(*xEntry) -= (*bInvAsEntry) * xAtLV;
	//Row LV
	(*xEntry) /= rateDecreaseLV;
	
	//Rows LV+1:End
	for(++bInvAsEntry, ++xEntry; bInvAsEntry != rateDecrease.end(); ++bInvAsEntry, xEntry++)
		(*xEntry) -= (*bInvAsEntry) * xAtLV;
}

void Simplex::updateData(unsigned enteringVariable, unsigned leavingVariable, vector<double>& costs)
{
	//Fix data structures after update
	constraints.at(enteringVariable).swap(constraints.at(leavingVariable));
	//Swap basis columns

	//If lv is real, set inv[vars[lv]] to ev.
	//set inv[vars[ev]] to lv.
	invVariables.at(variables.at(enteringVariable)) = leavingVariable;
	swap(variables.at(leavingVariable), variables.at(enteringVariable));
	//Keep track of variables
	swap(costs.at(leavingVariable), costs.at(enteringVariable));
	//Keep costs pointing at the right variables
	if(variables.at(enteringVariable) < 0) {
		//remove artificial variables from the problem when they leave the basis
		constraints.at(enteringVariable).swap(constraints.back());
		constraints.pop_back();
		swap(costs.at(enteringVariable), costs.back());
		costs.pop_back();
		if(variables.back() >= 0) invVariables.at(variables.back()) = enteringVariable;
		swap(variables.at(enteringVariable), variables.back());
		variables.pop_back();
		
	} else {
		invVariables.at(variables.at(enteringVariable)) = enteringVariable;
		//enteringVariable has been updated - this keeps track of LV of we need to.
	}
}

//TODO:
//GE/LE constraints: If we're satisfied, the artificial need not be artificial.
//EQ: If we're satisfied there need be no artificial at all.
void Simplex::addGEConstraint(unsigned var, double rhs)
{
	//TODO: check whether 0 <= var <= totalVariables?
	//TODO: fix the copy/paste job
	done = false;

	//Change the constraint matrix:
	for(vector<vector<double> >::iterator i = constraints.begin(); i != constraints.end(); ++i)
		i->push_back(0);
	constraints.at(invVariables.at(var)).back() = 1;

	//Add surplus variable
	constraints.push_back(vector<double>(x.size()+1, 0));
	constraints.back().back() = -1;
	variables.push_back(totalVariables++);
	costs.push_back(0);
	
	invVariables.push_back(constraints.size()-1);

	if(invVariables.at(var) < x.size())
		x.push_back(abs(x.at(invVariables.at(var)) - rhs));
	else x.push_back(rhs);
	
	//Enlarge inverseB
	for(vector<vector<double> >::iterator i = inverseB.begin(); i != inverseB.end(); ++i)
		i->push_back(0);
	inverseB.push_back(vector<double>(x.size(), 0));
	
	if((invVariables.at(var) >= x.size()-1 && 0 < rhs) || (invVariables.at(var) < x.size()-1 && x.at(invVariables.at(var)) < rhs)) {
		//Don't satisfy the constraint. Add an artificial.
		constraints.push_back(vector<double>(x.size(), 0));
		constraints.back().back() = 1;
		variables.push_back(--edgeArtificial);
		optimal = false;
	} else {
		//update invVariables, TODO: Check this
		invVariables.back() = x.size()-1;//?
		invVariables.at(variables.at(x.size()-1)) = constraints.size()-1;//?
	}
	swap(variables.back(), variables.at(x.size()-1));
	if(variables.back() > 0) invVariables.at(variables.back()) = variables.size()-1;

	constraints.at(x.size()-1).swap(constraints.back());
	int sign = inverseB.back().back() = constraints.at(x.size()-1).back();
	//corner of inverseB depends on whether it's the surplus or artificial
	if(invVariables.at(var) < inverseB.size()) {
		//var is basic, need to change inverseB
		vector<double>::const_iterator itsRow = inverseB.at(invVariables.at(var)).begin();
		for(vector<double>::iterator lastRow = inverseB.back().begin(); lastRow != inverseB.back().end(); ++lastRow, ++itsRow)
			*lastRow += sign * (*itsRow)*-1;
	}
}

void Simplex::addLEConstraint(unsigned var, double rhs)
{
	done = false;

	//Change the constraint matrix:
	for(vector<vector<double> >::iterator i = constraints.begin(); i != constraints.end(); ++i)
		i->push_back(0);
	constraints.at(invVariables.at(var)).back() = 1;

	//Add slack variable
	constraints.push_back(vector<double>(x.size()+1, 0));
	constraints.back().back() = 1;
	variables.push_back(totalVariables++);
	costs.push_back(0);
	
	invVariables.push_back(constraints.size()-1);

	if(invVariables.at(var) < x.size())
		x.push_back(abs(x.at(invVariables.at(var)) - rhs));
	else x.push_back(rhs);
	
	//Enlarge inverseB
	for(vector<vector<double> >::iterator i = inverseB.begin(); i != inverseB.end(); ++i)
		i->push_back(0);
	inverseB.push_back(vector<double>(x.size(), 0));
	
	if((invVariables.at(var) >= x.size()-1 && 0 > rhs) || (invVariables.at(var) < x.size()-1 && x.at(invVariables.at(var)) > rhs)) {
		//Don't satisfy the constraint. Add an artificial.
		constraints.push_back(vector<double>(x.size(), 0));
		constraints.back().back() = -1;
		variables.push_back(--edgeArtificial);
		optimal = false;
	} else {
		//update invVariables, TODO: Check this
		invVariables.back() = x.size()-1;//?
		invVariables.at(variables.at(x.size()-1)) = constraints.size()-1;//?
	}
	swap(variables.back(), variables.at(x.size()-1));
	if(variables.back() > 0) invVariables.at(variables.back()) = variables.size()-1;

	constraints.at(x.size()-1).swap(constraints.back());
	int sign = inverseB.back().back() = constraints.at(x.size()-1).back();
	//corner of inverseB depends on whether it's the slack or artificial
	if(invVariables.at(var) < inverseB.size()) {
		//var is basic, need to change inverseB
		vector<double>::const_iterator itsRow = inverseB.at(invVariables.at(var)).begin();
		for(vector<double>::iterator lastRow = inverseB.back().begin(); lastRow != inverseB.back().end(); ++lastRow, ++itsRow)
			*lastRow += sign * (*itsRow)*-1;
	}
}
